import numpy as np
import subprocess as sp
import sys
import os
from scipy.ndimage import binary_dilation,median_filter

### astronomy packages
from pyraf import iraf
from astropysics import coords
import pywcs
import pyfits
import cosmology as cosmos

def rescale_rexc(r_exc,z_low,z_high):
    d_low = cosmos.angular_distance(z_low)
    d_high = cosmos.angular_distance(z_high)    
    return r_exc*(d_low/d_high)
    
def dist2(x1,y1,x2,y2):
    "Computes the distance between two points with coordinates (x1,y1) and (x2,y2)"
    return ((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2))


def sex_to_degrees(ra,dec):
    ra=coords.AngularCoordinate(ra,sghms=True)
    dec=coords.AngularCoordinate(dec,sghms=False)
    return ra.degrees,dec.degrees

#==============================================================================
#  IMAGE Handling
#==============================================================================

def get_center_coords(imgname,ra,dec):
    """ Function ot convert from sky coordinates (RA,DEC) into image coordinates
    (XC,YC).
    """
    hdu=pyfits.open(imgname)
    wcs=pywcs.WCS(hdu[0].header)

    ctype=hdu[0].header["ctype1"]
    
    if 'RA' in ctype:
        sky=np.array([[ra,dec]],np.float_)
    else:
        sky=np.array([[dec,ra]],np.float_)

    pixcrd=wcs.wcs_sky2pix(sky,1)

    xc=pixcrd[0,0]
    yc=pixcrd[0,1]

    return xc,yc

def imagecut(imgname,ra,dec,path,exptime,hsize):

    hdu=pyfits.open(imgname)
    xc,yc=get_center_coords(imgname,ra,dec)

    nmgy=hdu[0].header['nmgy']
    xmax=hdu[0].header["naxis1"]
    ymax=hdu[0].header["naxis2"]
    
    
    if xc < 0 or xc > xmax:
        raise ValueError("Ra or Dec out of image range")
    if yc < 0 or yc > ymax:
        raise ValueError("Ra or Dec out of image range")
    
    if xc-hsize<=0:
        xc=hsize+1
    if yc-hsize<=0:
        yc=hsize+1
    
    if xc+hsize>xmax:
        xc=xmax-hsize-1
    if yc+hsize>ymax:
        yc=ymax-hsize-1

    xl=int(xc-hsize)
    xr=int(xc+hsize)
    yd=int(yc-hsize)
    yu=int(yc+hsize)

    if os.path.isfile(path+'galaxy.fits'):
        print "galaxy.fits exists, removing old file\n"
        os.system('rm '+path+'galaxy.fits')

    if os.path.isfile(path+'intermed.fits'):
        os.system('rm '+path+'intermed.fits')

    iraf.imcopy("%s[0]"%imgname,path+"intermed.fits")
    iraf.imarith("%sintermed.fits[%i:%i,%i:%i]"%(path,xl,xr,yd,yu),'*',(exptime/nmgy),path+"galaxy.fits")
    os.system('rm %sintermed.fits'%path)


    return xc,yc


#==============================================================================
# HEADER handling
#==============================================================================

def get_ids(imgfile):
    fname=imgfile.split('/')[-1]                            #split string by / to select the frame-******-*-****.fits full name
    parts=fname.split('-')                                  #split string by - to separete the numbers of the name of the file
    run=int(parts[2])                                       #set run to the 3rd number of the previous split 
    camcol=int(parts[3])                                    #set camcol to the 4th number of the previous split 
    field=int(parts[-1].split('.')[0])                      #set field to the last number of previous split (additional splitting is required to separate the .fits termination)
    return run,camcol,field
    
def header_keys(path,imgfile,band):
    run,camcol,field=get_ids(imgfile)                                       #get the numbers run, camcol and field for the frame image
    hdu=pyfits.open(path+'/photoField-%06i-%i.fits'%(run,camcol))            #open the photoField image using the retrieved numbers
    table=hdu[1].data                                                       #define the data table as the second extension of the fits file
    FIELDS=table.field('field')                                             #define the subtable containg the field numbers
    gain=table.field('gain')[FIELDS==field][0,int(band)]                  
    rdnoise=table.field('dark_variance')[FIELDS==field][0,int(band)]      #select subtables containg keywords of interest and selecting the values
    airmass=table.field('airmass')[FIELDS==field][0,int(band)]            # for the field of the frame image and the band ('ugriz'<->'12345')
    sky_nmgy=table.field('sky_frames_sub')[FIELDS==field][0,int(band)]    #check http://data.sdss3.org/datamodel/files/BOSS_PHOTOOBJ/RERUN/RUN/photoField.html
#    print gain,rdnoise,airmass,sky_nmgy,'r'
    return gain,rdnoise,airmass,sky_nmgy                                    # for more details on all the keywords available

def update_header(path,imgfile,imgname,exptime=0.0,band=2):
    gain,rdnoise,airmass,sky_nmgy=header_keys(path,imgfile,band=band)       #get keywords values from photoField
    filters=['u','g','r','i','z']                                           #set the filters list of the SDSS
    hdu=pyfits.open(imgname,mode='update')                                  #open frame image to update
    hdr=hdu[0].header                                                       #set header variable
    if exptime!=0:
#    hdr.update('OLD_EXPT',hdr['EXPTIME'],'Old Exposure Time Value',after='EXPTIME') #copy EXPTIME keyword to OLD_EXPT
        hdr.update('EXPTIME',exptime,'Exposure Time of the High Redshift Field')   #set new EXPTIME = 1.0 to avoid miscalculated magnitudes
    hdr.update('GAIN',gain,'Gain value from photoField')                            #create gain keyword
    hdr.update('RDNOISE',rdnoise,'Dark_Variance value from photoField')             #create rdnoise keyword
    hdr.update('AIRMASS',airmass,'Airmass value from photoField')                   #create airmass keyword
    hdr.update('FILTERS',filters[band],'Filter of this image')                    #create filters keyword
    hdr.update('SKY_NMGY',sky_nmgy,'Sky value (in nmgy) obtained from photoField')  #create sky_nmgy keyword
    hdu.close()                                                                     #close file to save changes
    return None


#==============================================================================
#  Create PSF from SDSS psField
#==============================================================================

def prep_psf(psfield,xc,yc,path,band='g'):
    band_dict={'u':'1','g':'2','r':'3','i':'4','z':'5'}    
    
    bas=open(path+'psf.sh','w')
    idlcommand='psf_creator,"'+psfield+'",%i,%i,%s\n'%(xc,yc,band_dict[band])
    bas.write('#!/bin/bash\n')
    bas.write('idl << EOF\n')
    bas.write('\t.r psf\n')
    bas.write('\t'+idlcommand)
    bas.write('EOF')
    bas.close()
    os.system('chmod +x '+path+'psf.sh')
    os.chdir(path)
    os.popen('./psf.sh')
    os.chdir('..')
    return 'psf%s.fits'%(band_dict[band])

#==============================================================================
#  SExtractor
#==============================================================================

def gen_segmap_sex(fname,zeropoint,thresh=3.0,pix_scale=0.03,seeing=0.09,weights='none'):
    root=os.getcwd()
    "Returns the segmentation map generated from SExtractor"
    
    if 'linux' in sys.platform:
        sexcommand='sextractor'
    elif 'darwin' in sys.platform:
        sexcommand='sex'
    else:
        sexcommand='sextractor'
        
    if 'none' in weights:
        sp.call("%s %s -MAG_ZEROPOINT %.8f -CATALOG_NAME temp.cat -CHECKIMAGE_TYPE SEGMENTATION -CHECKIMAGE_NAME %s/map.fits -DETECT_MINAREA 5 -FILTER Y -DETECT_THRESH %f -ANALYSIS_THRESH %f -PIXEL_SCALE %.2f -SEEING_FWHM %.2f -WEIGHT_TYPE NONE "%(sexcommand,fname,zeropoint,root,thresh,thresh,pix_scale,seeing),shell=True)
        print("%s %s -MAG_ZEROPOINT %.8f -CATALOG_NAME temp.cat -CHECKIMAGE_TYPE SEGMENTATION -CHECKIMAGE_NAME %s/map.fits -DETECT_MINAREA 5 -FILTER Y -DETECT_THRESH %f -ANALYSIS_THRESH %f -PIXEL_SCALE %.2f -SEEING_FWHM %.2f -WEIGHT_TYPE NONE "%(sexcommand,fname,zeropoint,root,thresh,thresh,pix_scale,seeing))
    else:
        sp.call("%s %s -MAG_ZEROPOINT %.8f -CATALOG_NAME temp.cat -CHECKIMAGE_TYPE SEGMENTATION -CHECKIMAGE_NAME %s/map.fits -DETECT_MINAREA 5 -FILTER Y -DETECT_THRESH %f -ANALYSIS_THRESH %f -PIXEL_SCALE %.2f -SEEING_FWHM %.2f -WEIGHT_TYPE MAP_RMS -WEIGHT_IMAGE %s"%(sexcommand,fname,zeropoint,root,thresh,thresh,pix_scale,seeing,weights),shell=True)
    segmap=pyfits.getdata('%s/map.fits'%root)
    sp.call('rm %s/map.fits'%(root),shell=True)
    return segmap

def get_sex_pars(xc,yc,rmax,catfile='test.cat',psf=False):
    """Returns the SExtractor parameters from the catalog associated with the
    segmentation map for the source closest to the VUDS catalog coordinates."""

    f=open(catfile,'r');txt=f.readlines();f.close()
    last_line=txt[-1]
    if last_line[0]=='#':        
        return -99,-99,-99,-99,-99,-99,-99,-99,-99,-99

    if psf:
        xs,ys,mag,mum,re,a,kr,e,t,cs,xp,yp=np.loadtxt(catfile,unpack=True,ndmin=2)
    else:
        xs,ys,mag,mum,re,a,kr,e,t,cs,xp,yp,isoarea=np.loadtxt(catfile,unpack=True,ndmin=2)

    n=a*kr/re
    e=1-e
    t=t
##    if np.size(xs)==1:
##        separation=np.sqrt(dist2(xs,ys,xc,yc))
##        if psf:
##            return xp,yp
##        else:
##            return [xs],[ys],[mag],[re],[n],[e],[t],0,separation,[isoarea]
##    else:
    dists=(dist2(xs,ys,xc,yc))
    obj_num=np.where(dists == min(dists))
    if min(dists)>rmax*rmax:
        return -99
    if psf:
        return xp,yp,obj_num
    else:
        return xs,ys,mag,re,n,e,t,obj_num,np.sqrt(min(dists)),isoarea

#==============================================================================
# MASK preparation
#==============================================================================

def sky_value(img,k=3):
    """Compute the value of sky background and respective standard
    deviation using a sigma clipping method."""
    media=np.nanmean(img)
    dev=np.nanstd(img)

    back=img.copy()
    back=back[back!=0]
    thresh=media+k*dev
    npix = len(img[abs(img)>=thresh])
    while npix>0:
        back = back[abs(back)<thresh]
        media=np.nanmean(back)
        dev=np.nanstd(back)
        thresh=media+k*dev
        npix = len(back[abs(back)>=thresh])
    return media,dev

def outlier_detection(img,k=15,s=1,mask=None):
    """ Flags the pixels that are recognized as outliers from a median_filter passage.
    """
    N,M=img.shape
    out_map=np.zeros(img.shape)
    blurred = median_filter(img,size=s+2)
    diff=img-blurred

    out_map=np.zeros(img.shape)
    ks=np.where(abs(diff)>np.mean(diff)+k*np.std(diff))  

    if mask is None:
        out_map[ks]=1        
    else:
        out_map[ks]=1-mask[ks]
        
#    fig,ax=subplots(1,4)
#    ax[0].imshow(img)
#    ax[1].imshow(blurred)
#    ax[2].imshow(diff)
#    ax[3].imshow(out_map)
#    show()
    return out_map

def distance_matrix(xc,yc,img):
    """Compute a matrix with distances and a sorted array with unique distance
    values from each pixel position to xc,yc"""
    N,M=img.shape
    Dmat=np.zeros([N,M])
    for i in range(N):
        for j in range(M):
            Dmat[i,j] = dist2(xc,yc,i,j)
    Dmat=np.sqrt(Dmat)
    dists=np.sort(np.array(list(set(Dmat.reshape(np.size(Dmat))))))
    return Dmat, dists
    
def select_object_map(xc,yc,segmap,pixscale,rexclusive):
    """Select the segmentation map of the object by settin all regions within
    approximantely 1 arcsecond to be assigned to the same object.
    """
    Dmat,d=distance_matrix(yc,xc,segmap)    
    

    
    s_values = set(segmap[Dmat<rexclusive/pixscale])
    new_map = segmap.copy()    
    
    for s in s_values:
        if s==0:
            continue
        new_map[new_map==s]=-1
    new_map[new_map>0]=0
    
#    import matplotlib.pyplot as mpl
#    from matplotlib.patches import Rectangle,Ellipse,Circle,Polygon
#    print rexclusive,rexclusive/pixscale
#    fig,ax=mpl.subplots(1,2)
#    ax[0].imshow(segmap)
#    C = Circle((xc,yc),radius=rexclusive/pixscale,color='red',fill=False)
#    ax[0].add_artist(C)
#    ax[1].imshow(-new_map)
#    mpl.show()
    
    return -new_map

def map_to_mask(img,segmap,xc,yc,maskname,kout,pixscale,rexc,s=1):
    """Converts the segmentation map from Sextractor into a mask file for GALFIT.
    Adds outlier pixels to the map and dilates the map to cover faint surface brightness
    regions."""
    segmap=segmap.copy()
    omap=outlier_detection(img,kout,s=s,mask=segmap)
#    imshow(omap);show()
    
    segmap[omap==1]=-1
    new_map=select_object_map(xc,yc,segmap,pixscale=pixscale,rexclusive=rexc)

#    imshow(new_map);show()

#    print len(segmap[segmap==-1])

    segmap[new_map==1]=0        ## deselecting object
    segmap[segmap!=0]=1         ## setting all segmap values to one
    segmap[img==0]=1            ## adding pixels where image as zero value
    segmap[segmap==-1]=1        ## adding pixel outliers that fall outside de fitting region
     
    SIZE=6
    STRUCTURE = np.zeros([SIZE,SIZE])
    dmat,d= distance_matrix(SIZE/2.-0.5,SIZE/2.-0.5,STRUCTURE)
    STRUCTURE[dmat<SIZE/2]=1
    segmap = binary_dilation(segmap,structure=STRUCTURE).astype(int)
     
    pyfits.writeto(maskname,segmap,clobber=True)
    return segmap

#==============================================================================
# GALFIT file creation
#==============================================================================

def write_object(f,model,x,y,m,re,n,ba,pa,num,fixpars=None):
    if fixpars==None:
        fixpars={'x':1,'y':1,'m':1,'re':1,'n':1,'q':1,'pa':1}
    f.write("#Object number: %i\n"%num)    
    f.write(' 0) %s             # Object type\n'%model)
    f.write(' 1) %6.4f %6.4f  %i %i    # position x, y        [pixel]\n'%(x,y,fixpars['x'],fixpars['y']))
    f.write(' 3) %4.4f      %i       # total magnitude\n' %(m,fixpars['m']))
    f.write(' 4) %4.4f       %i       #     R_e              [Pixels]\n'%(re,fixpars['re']))
    f.write(' 5) %4.4f       %i       # Sersic exponent (deVauc=4, expdisk=1)\n'%(n,fixpars['n']))  
    f.write(' 9) %4.4f       %i       # axis ratio (b/a)   \n'%(ba,fixpars['q']))
    f.write('10) %4.4f       %i       # position angle (PA)  [Degrees: Up=0, Left=90]\n'%(pa,fixpars['pa']))
    f.write(' Z) 0                  #  Skip this model in output image?  (yes=1, no=0)\n')
    f.write(' \n')
    return

def galfit_input_file(f,magzpt,sky,xsize,ysize,sconvbox,pixscale,imgname='galaxy.fits',outname="results.fits",psfname='psf.fits',maskname="none",signame='none',constname='none',fixpars=None):
    if fixpars==None:
        fixpars={'sky':1}
    f.write("================================================================================\n")
    f.write("# IMAGE and GALFIT CONTROL PARAMETERS\n")
    f.write("A) %s         # Input data image (FITS file)\n"%imgname)
    f.write("B) %s        # Output data image block\n"%outname)
    f.write("C) %s                # Sigma image name (made from data if blank or 'none' \n"%signame)
    f.write("D) %s         # Input PSF image and (optional) diffusion kernel\n"%psfname)
    f.write("E) 1                   # PSF fine sampling factor relative to data \n")
    f.write("F) %s                # Bad pixel mask (FITS image or ASCII coord list)\n"%maskname)
    f.write("G) %s                # File with parameter constraints (ASCII file) \n"%constname)
    f.write("H) 1    %i   1    %i # Image region to fit (xmin xmax ymin ymax)\n"%(xsize+1,ysize+1))
    f.write("I) %i    %i          # Size of the convolution box (x y)\n"%(sconvbox,sconvbox))
    f.write("J) %7.5f             # Magnitude photometric zeropoint \n"%magzpt)
    f.write("K) %.3f %.3f        # Plate scale (dx dy)   [arcsec per pixel]\n"%(pixscale,pixscale))
    f.write("O) regular             # Display type (regular, curses, both)\n")
    f.write("P) 0                   # Options: 0=normal run; 1,2=make model/imgblock and quit\n")
    f.write("\n")
    f.write("# INITIAL FITTING PARAMETERS\n")
    f.write("#\n")
    f.write("#For object type, the allowed functions are:\n")
    f.write("#nuker, sersic, expdisk, devauc, king, psf, gaussian, moffat,\n")
    f.write("#ferrer, and sky.\n")
    f.write("#\n")
    f.write("#Hidden parameters will only appear when theyre specified:\n")
    f.write("#C0 (diskyness/boxyness),\n")
    f.write("#Fn (n=integer, Azimuthal Fourier Modes).\n")
    f.write("#R0-R10 (PA rotation, for creating spiral structures).\n")
    f.write("#\n")
    f.write("# ------------------------------------------------------------------------------\n")
    f.write("#  par)    par value(s)    fit toggle(s)   parameter description\n")
    f.write("# ------------------------------------------------------------------------------\n")
    f.write("\n")

    obj=open('galfit_object.temp','r')
    objects=obj.readlines()
    for line in objects:
        f.write(line)
    obj.close()
    
    f.write("# Object: Sky\n")
    f.write(" 0) sky                    #  object type\n")
    f.write(" 1) %7.4f      %i          #  sky background at center of fitting region [ADUs]\n"%(sky,fixpars['sky']))
    f.write(" 2) 0.0000      0          #  dsky/dx (sky gradient in x)\n")
    f.write(" 3) 0.0000      0          #  dsky/dy (sky gradient in y)\n")
    f.write(" Z) 0                      #  output option (0 = resid., 1 = Dont subtract)")    
    f.close()
    return

#==============================================================================
#  READ GALFIT results
#==============================================================================

def read_results_file(fname):
    try:
        hdu=pyfits.open(fname)
        chi=hdu[2].header['CHI2NU']
        xc=hdu[2].header['1_XC'].translate(None,'[*').replace('+/-','').replace(']',' -99.00')
        yc=hdu[2].header['1_YC'].translate(None,'[*').replace('+/-','').replace(']',' -99.00')
        M=hdu[2].header['1_MAG'].translate(None,'[*').replace('+/-','').replace(']',' -99.00')
        R=hdu[2].header['1_RE'].translate(None,'[*').replace('+/-','').replace(']',' -99.00')
        N=hdu[2].header['1_N'].translate(None,'[*').replace('+/-','').replace(']',' -99.00')
        Q=hdu[2].header['1_AR'].translate(None,'[*').replace('+/-','').replace(']',' -99.00')
        T=hdu[2].header['1_PA'].translate(None,'[*').replace('+/-','').replace(']',' -99.00')
        F= hdu[2].header['FLAGS'].replace(' ',',')
        if ('1' in F.split(',')):
            F=1
        elif ('2' in F.split(',')):
            F=2
        else:
            F=0
    except IOError:
        xc,yc,M,R,N,Q,T,chi,F="-99 -99","-99 -99","-99 -99","-99 -99","-99 -99","-99 -99","-99 -99","-99",-9
    return xc,yc,M,R,N,Q,T,chi,F



import .auxiliary_functions_sdss as afs
import warnings
import argparse
import .dopterian

def lum_evolution(zlow,zhigh):
    "Defined Luminosity evolution from L* of Sobral et al. 2013."
    def luminosity(z):
        logL = 0.45*z+41.87
        return 10**(logL)
    return luminosity(zhigh)/luminosity(zlow)

warnings.simplefilter('ignore')

parser = argparse.ArgumentParser(description="Code to download images of galaxies from catalog form SDSS DR9 server.")
parser.add_argument('catalog',metavar='NAME',type=str,help="Catalog on which to run GALFIT. Format should be ID,RA,DEC,...")
parser.add_argument('-z','--redshifts',metavar='NUM',type=str,help="Redshift of the smoothed images (Only aplicable when using --smooth).")
parser.add_argument('-i','--name',metavar='NAME',type=str,help="Name of the first galaxy to start with")
parser.add_argument('--highz_sky',metavar='fname',type=str,default="none",help="Image on which dopterian drops the high z galaxy.")
parser.add_argument('--highz_expt',metavar='pixel_scale',type=float,default=2028.,help="Pixel scale for the high redshift imaging.")
parser.add_argument('--highz_zp',metavar='zeropoint',type=float,default=25.947,help="Zeropoint for the high redshift imaging.")
parser.add_argument('--highz_psf',metavar='fname',type=str,help="PSF for the high redshift imaging.")
parser.add_argument('--ferengi',metavar='MODE',type=str,choices=('all','dimming','mag','none'),help="Options to run ferengi: with dimming and mag evolution, justo one of them or none.")
parser.add_argument('-b','--band',metavar='NAME',type=str,default='g',help="Band where to run ferengi. Default:g")
parser.add_argument('-t','--threshold',metavar='NAME',type=float,default=3.0,help="Threshold for SExtractor detections.")
parser.add_argument('-r','--rexclusive',metavar='NAME',type=float,default=20,help="Exlcusion radius for masking. Objects falling inside the given radius are not masked.")
parser.add_argument('-S','--seeing',metavar='NAME',type=float,default=0.09,help="Typical seeing of the images.")
parser.add_argument('--lowz_pixscale',metavar='pix',type=float,default=0.396,help="Pixelscale for the low redshift imaging.")
parser.add_argument('--highz_pixscale',metavar='pix',type=float,default=0.03,help="Pixelscale for the high redshift imaging.")

NOIMAGE=[]
band_dict={'u':'1','g':'2','r':'3','i':'4','z':'5'}
central_bands = {'u':3561,'g':4718,'r':6185,'i':7501,'z':8962}

if __name__=='__main__':
    args = parser.parse_args()

    SDSS_BANDS='ugriz'
    SDSS_PLATE=args.lowz_pixscale
    PS_HIGHZ=args.highz_pixscale ## pixscale high z imaging
    EXPTIME_SMOOTH=args.highz_expt # SDSS EXPTIME in seconds by default

    root=afs.os.getcwd()
    catalog_name=args.catalog
    startid=args.name
    threshold=args.threshold
    Rexclusion = args.rexclusive
    band=args.band
    SEEING=args.seeing
    MZPT=args.highz_zp
    redshifts = [float(z) for z in args.redshifts.split(',')]


    prefix = args.ferengi


    print("""
    GENERAL CONDITIONS FOR FERENGI and then GALFIT
    ----------------------------------------------------------------------------------
    Running DOPTERIAN on catalog %s and photometric bands %s
    Saving results to file %s
    
    DOPTERIAN mode = %s
    
    Pixel scale (highz) = %.3f
    Typical seeing (highz) = %.2f
    Mag. zeropoint (highz) = %.5f
    Exp. time (highz) = %.5f

    Pixel scale (lowz) = %.3f
    Mag. zeropoint (lowz) = Invidual (from image header)
    
    PSF image name: %s
    SKY image name: %s
    
    Detection threshold = %.1f sigma
    
    Start ID = %s
    ----------------------------------------------------------------------------------
    """%(catalog_name,band,'dopterian_%s_simulation_results_X.XXX.txt'%(args.ferengi),args.ferengi,PS_HIGHZ,SEEING,MZPT,EXPTIME_SMOOTH,args.lowz_pixscale,args.highz_psf,args.highz_sky,threshold,startid))

##    question=raw_input("Are these values ok? Do you want to continue? (y/n)\t: ")
##
##    if question=='n' or question=='no':
##        print "Now exiting program!"
##        afs.sys.exit()
##    else:
##        pass


    SExcatalog=[open('%s/dopterian_%s_SExcatalog_%.3f.txt'%(root,prefix,z),'w') for z in redshifts]
    [SExcatalog[k].write('# ID\tx\ty\tmag\tre\tn\tb/a\tPA\tSeparation\n') for k in range(len(redshifts))]

    table=[open('%s/dopterian_%s_simulation_results_%.3f.txt'%(root,prefix,z),'w') for z in redshifts]
    [table[k].write('# ID\tx\tx_err\ty\ty_err\tmag\tmag_err\tre\tre_err\tn\tn_err\tb/a\tb/a_err\tPA\tPA_err\tchi\tflag\tSeparation\n') for k in range(len(redshifts))]


    Info_Table=afs.np.loadtxt(catalog_name,dtype={'names':('ID','RA','DEC','Z'),'formats':('S50','S25','S25','f4')})
    ngalaxies = len(Info_Table['ID'])    
######

    start=0

    for i in range(ngalaxies):

        if (not Info_Table['ID'][i]==startid) and (start==0) and (startid!=None):
            continue
        else:
            start=1
        
        path="./%s"%(Info_Table['ID'][i])

        if not afs.os.path.isdir(path):
            print("\033[01;31m{0}\033[00m".format("Directory not found for galaxy %s"%Info_Table['ID'][i]))
            NOIMAGE.append(Info_Table['ID'][i])
            continue
        elif afs.os.path.isfile("%s/galaxy.fits"%path):
            input_name = "%s/galaxy.fits"%path
            psf_name = afs.sp.check_output('ls %s/psf%s.fits'%(path,band_dict[band]),shell=True).split()[0]
        else:
            print("\033[01;31m{0}\033[00m".format("galaxy.fits not found for galaxy %s"%Info_Table['ID'][i]))
            NOIMAGE.append(Info_Table['ID'][i])

        ra,dec=afs.sex_to_degrees(Info_Table['RA'][i],Info_Table['DEC'][i])

        hdu=afs.pyfits.open(input_name,mode='update')
        ximg,yimg=afs.get_center_coords(input_name,ra,dec)
        Nori,Mori = hdu[0].data.shape
        nmgy=hdu[0].header['nmgy']  #nanomaggies per count
        exptime=float(hdu[0].header['exptime'])
        zp_sdss=22.5-2.5*afs.np.log10(nmgy)     #mzp - 22.5 = -2.5log(nmgy/1)  calibration 1 nmgy -> 22.5 mag
        hdu.close()
        
        lowz_info = {'redshift':Info_Table['Z'][i],'psf':psf_name,'zp':zp_sdss,'exptime':exptime,'filter':band,'lam_eff':central_bands[band],'pixscale':SDSS_PLATE}
        highz_info = {'redshift':0.8,'psf':args.highz_psf,'zp':args.highz_zp,'exptime':args.highz_expt,'filter':'f814w','lam_eff':8140.0,'pixscale':PS_HIGHZ}

        for k in range(len(redshifts)):
            highz=float(redshifts[k])
            highz_info['redshift']=highz

            if args.ferengi=='all':
                dopterian.ferengi(input_name,args.highz_sky,lowz_info,highz_info,['%s/stamp_z%.3f.fits'%(path,highz),'%s/psf_z%.3f.fits'%(path,highz)],evo=lum_evolution,noflux=False)
            elif args.ferengi=='mag':
                dopterian.ferengi(input_name,args.highz_sky,lowz_info,highz_info,['%s/stamp_z%.3f.fits'%(path,highz),'%s/psf_z%.3f.fits'%(path,highz)],evo=lum_evolution,noflux=True)
            elif args.ferengi=='dimming':
                dopterian.ferengi(input_name,args.highz_sky,lowz_info,highz_info,['%s/stamp_z%.3f.fits'%(path,highz),'%s/psf_z%.3f.fits'%(path,highz)],evo=None,noflux=False)
            elif args.ferengi=='none':
                dopterian.ferengi(input_name,args.highz_sky,lowz_info,highz_info,['%s/stamp_z%.3f.fits'%(path,highz),'%s/psf_z%.3f.fits'%(path,highz)],evo=None,noflux=True)
            else:
                raise ValueError('Invalid mode selected: %s'%args.ferengi)

####################################################################################################################################################
####################################################################################################################################################
            imgname='%s/dopterian_galaxy_z%.3f_%s.fits'%(path,highz,args.ferengi)
            if afs.os.path.isfile('%s/stamp_z%.3f.fits'%(path,highz)):
                afs.iraf.imarith('%s/stamp_z%.3f.fits'%(path,highz),'*',highz_info['exptime'],imgname)
                afs.iraf.hedit(imgname,'EXPTIME',highz_info['exptime'],addonly='yes',verify='no',show='yes')
                stamp_red = afs.pyfits.getdata(imgname)
            else:
                table[k].write('%s\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\n'%(Info_Table['ID'][i]))
                SExcatalog[k].write('%s\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\n'%(Info_Table['ID'][i],-99,-99,-99,-99,-99,-99,-99,-99))
                continue
                
            N,M=stamp_red.shape
            assert N==M
            xmed = ximg*(float(N)/Nori)
            ymed = yimg*(float(M)/Mori)
            
            segmap = afs.gen_segmap_sex(imgname,highz_info['zp'],pix_scale=highz_info['pixscale'],thresh=threshold,seeing=SEEING)
            afs.map_to_mask(stamp_red,segmap,xmed,ymed,'%s/mask_z%.3f.fits'%(path,highz),10,highz_info['pixscale'],afs.rescale_rexc(Rexclusion,lowz_info['redshift'],highz_info['redshift']))
            xs,ys,mag,re,n,e,t,num,sep,isoarea=afs.get_sex_pars(xmed,ymed,N,catfile='temp.cat')
            
            try:
                len(xs)
            except TypeError:
                xs,ys,mag,re,n,e,t=[xs],[ys],[mag],[re],[n],[e],[t]
                num=0
            mcorr=2.5*afs.np.log10(highz_info['exptime'])


            SExcatalog[k].write('%s\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\t%10.4f\n'%(Info_Table['ID'][i],xs[num],ys[num],mag[num]+mcorr,re[num],n[num],e[num],t[num],sep))
            print 'SEx Pars',xs[num],ys[num],mag[num]+mcorr,re[num],n[num],e[num],t[num]                                         
    
##            if sep >10:
##                print "No galaxy detected at X,Y=center,center"
##                table[k].write('%s\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\n'%(Info_Table['ID'][i]))
##                sp.call('rm %s/mask_z%.3f.fits'%(path,highz),shell=True)
##                k+=1
##                continue
    
            if sep==-99.0:
                print "Nothing detected in stamp image"
                table[k].write('%s\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\t-99.0000\n'%(Info_Table['ID'][i]))
                afs.sp.call('rm %s/mask_z%.3f.fits'%(path,highz),shell=True)
                continue
            
            
            f1=open('%s/galfit_object.temp'%(root),'w')              
            afs.write_object(f1,'sersic',xs[num],ys[num],mag[num]+mcorr,re[num],n[num],e[num],t[num],0)
            f1.close()
            afs.sp.call("rm temp.cat",shell=True,stderr=afs.sp.PIPE)

            
####################################################################################################################################################
####################################################################################################################################################
 

            f2=open('%s/galfit_file_z%.3f'%(root,highz),'w')
            afs.galfit_input_file(f2,highz_info['zp'],0.0,N,M,N,PS_HIGHZ,imgname=imgname,psfname='%s/psf_z%.3f.fits'%(path,highz),outname='%s/result_z%.3f.fits'%(path,highz),maskname='%s/mask_z%.3f.fits'%(path,highz))
            f2.close()
    
            afs.sp.call('galfit galfit_file_z%.3f >> galfit.log'%(highz),shell=True,stderr=afs.sp.PIPE)
#            sp.call('galfit galfit_file%i_z%.3f'%(runtype,highz),shell=True)
            X,Y,MAG,R,S,Q,T,CHI,F=afs.read_results_file('%s/result_z%.3f.fits'%(path,highz))
            print 'GALFIT Pars',X,Y,MAG,R,S,Q,T,CHI,F                                          

            table[k].write('%s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%i\t%3.2f\n'%(Info_Table['ID'][i],X,Y,MAG,R,S,Q,T,CHI,F,sep))
            
            if afs.os.path.isdir('%s/z%.2f'%(path,highz)):
                pass
            else:
                afs.sp.call('mkdir %s/z%.2f'%(path,highz),shell=True)
            afs.sp.call('cp %s %s/z%.2f'%(imgname,path,highz),shell=True)
            afs.sp.call('cp %s/galfit_file_z%.3f %s/z%.2f/dopterian_galfit_file_z%.3f '%(root,highz,path,highz,highz),shell=True)
            
        afs.sp.call('rm %s/dopterian*.fits %s/stamp_z*.fits %s/result_z*.fits %s/psf_z*.fits galfit_objects.temp galfit_file_z* galfit.* fit.log %s/mask_z*.fits'%(path,path,path,path,path),shell=True)

        print "--------------> Done with galaxy %s \n"%Info_Table['ID'][i]

    [SExcatalog[k1].close() for k1 in range(len(redshifts))]
    [table[k1].close() for k1 in range(len(redshifts))]



import numpy as np
import numpy.random as npr
import scipy.integrate as scint 
import scipy.optimize as scopt
import scipy.ndimage as scndi
import astropy.io.fits as pyfits
import matplotlib.pyplot as mpl
import .cosmology as cosmos
import astropy.modeling as apmodel
import astropy.convolution as apcon
import warnings

#==============================================================================
#  CONSTANTS
#==============================================================================
version = '1.0.0'

c = 299792458. ## speed of light


## SDSS maggies to lupton
magToLup = {'u':1.4e-10,'g':0.9e-10,'r':1.2e-10,'i':1.8e-10,'z':7.4e-10}

#==============================================================================
# HELPER FUNCTIONS
#==============================================================================

def nu2lam(nu):
    "From Hz to Angstrom"
    return c/nu*1e-10
    
def lam2nu(lam):
    "From Angstrom to Hz"
    return c/lam*1e10
    
def maggies2mags(maggies):
    return -2.5*np.log10(maggies)

def mags2maggies(mags):
    return 10**(-0.4*mags)

def maggies2fnu(maggies):
    "From maggies to units of [erg s-1 Hz-1 cm-2]"
    return 3531e-23*maggies 

def fnu2maggies(fnu):
    "From [erg s-1 Hz-1 cm-2] to maggies"
    return 3631e23*fnu

def fnu2flam(fnu,lam):
    return c*1e10/lam**2*fnu
    
def flam2fnu(flam,lam):
    return flam/c/1.0e10*lam**2

def lambda_eff(lam,trans):
    "Calculate the mean wavelength of a filter"
    indexs = np.where(lam != 0)[0]
    if len(indexs)==0:
        raise ValueError('ERROR: no non-zero wavelengths')
    else:
        Lambda=np.squeeze(lam[indexs])
        Transmission=np.squeeze(trans[indexs])
        return scint.simps(Lambda*Transmission,Lambda)/scint.simps(Transmission,Lambda)

def cts2maggies(cts,exptime,zp):
    return cts/exptime*10**(-0.4*zp)
    
def cts2mags(cts,exptime,zp):
    return maggies2mags(cts2maggies(cts,exptime,zp))

def maggies2cts(maggies,exptime,zp):
    return maggies*exptime/10**(-0.4*zp)
    
def mags2cts(mags,exptime,zp):
    return maggies2cts(mags2maggies(mags),exptime,zp)


def maggies2lup(maggies,filtro):
    b = magToLup[filtro]
    return -2.5/np.log(10)*(np.arcsinh(maggies/b*0.5)+np.log(b))
    
def lup2maggies(lup,filtro):
#    maggies=lup
    b = magToLup[filtro]
    return 2*b*np.sinh(-0.4*np.log(10)*lup-np.log(b))
    
def random_indices(size,indexs):
    "Returns an array of a set of indices from indexs with a size with no duplicates."
    return npr.choice(indexs,size=size,replace=False)

def edge_index(a,rx,ry):
    "The routine creates an index list of a ring with width 1 around the centre at radius rx and ry"
    N,M=a.shape
    XX,YY=np.meshgrid(np.arange(N),np.arange(M))
    
    Y = np.abs(XX-N/2.0).astype(np.int64)
    X = np.abs(YY-M/2.0).astype(np.int64)
    
    idx = np.where(((X==rx) * (Y<=ry)) + ((Y==ry) * (X<=rx)))
####    CHECK
    return idx


def dist_ellipse(img,xc,yc,q,ang):
    "Compute distance to the center xc,yc in elliptical apertures. Angle in degrees."
    ang=np.radians(ang)

    X,Y = np.meshgrid(range(img.shape[1]),range(int(img.shape[0])))
    rX=(X-xc)*np.cos(ang)-(Y-yc)*np.sin(ang)
    rY=(X-xc)*np.sin(ang)+(Y-yc)*np.cos(ang)
    dmat = np.sqrt(rX*rX+(1/(q*q))*rY*rY)
    return dmat

def resistent_mean(a,k):
    """Compute the mean value of an array using a k-sigma clipping method
    """
    
    media=np.nanmean(a)
    dev=np.nanstd(a)

    back=a.copy()
    back=back[back!=0]
    thresh=media+k*dev
    npix = len(a[a>=thresh])
    while npix>0:
        back = back[back<thresh]
        media=np.mean(back)
        dev=np.std(back)
        thresh=media+k*dev
        npix = len(back[back>=thresh])
        
    nrej = np.size(a[a>=thresh])
    return media,dev,nrej
  
def ring_sky(image,width0,nap,x=None,y=None,q=1,pa=0,rstart=None,nw=None):
    """ For an image measure the flux around postion x,y in rings with
    axis ratio q and position angle pa with nap apertures (for sky slope). rstart indicates 
    the starting radius and nw (if set) limits the width to the number of apertures
    and not calculated in pixels.
    """
    
    if nap<=3:
        raise ValueError('Number of apertures must be greater than 3.')

    if type(image)==str:
        image=pyfits.getdata(image)
    
    N,M=image.shape
    if rstart is None:
        rstart=0.05*min(N,M)
    
    if x is None and y is None:
        x=N*0.5
        y=M*0.5
    elif x is None or y is None:
        raise ValueError('X and Y must both be set to a value')
    else:
        pass
    
    rad = dist_ellipse(image,x,y,q,pa)
    max_rad=0.95*np.amax(rad)
    
    if nw is None:
        width=width0
    else:
        width=max_rad/float(width0)
    
    media,sig,nrej=resistent_mean(image,3)
    sig*=np.sqrt(np.size(image)-1-nrej)
    
    if rstart is None:
        rhi=width
    else:
        rhi=rstart
    
    nmeasures=2
    
    r=np.array([])
    flux=np.array([])
    i=0
    while rhi<=max_rad:
        extra=0
        ct=0
        while ct<10:
            idx = (rad<=rhi+extra)*(rad>=rhi-extra)*(np.abs(image)<3*sig)
            ct=np.size(image[idx])
            
            extra+=1
            if extra>max(N,M)*2:
                break
            
        if ct<5:
            sky = flux[len(flux)-1]
        else:
            sky = resistent_mean(image[idx],3)[0]
        
        r=np.append(r,rhi-0.5*width)
        flux=np.append(flux,sky)
        
        i+=1
        if np.size(flux) > nap:
            pars,err = scopt.curve_fit(lambda x,a,b:a*x+b,r[i-nap+1:i],flux[i-nap+1:i])
            slope=pars[0]
            if slope>0 and nmeasures==0:
                break
            elif slope>0:
                nmeasures-=1
            else:
                pass

        rhi += width
    sky = resistent_mean(flux[i-nap+1:i],3)[0]        
    return sky

def ferengi_make_psf_same(psf1,psf2):
    "Compares the size of both psf images and zero-pads the smallest one so that they have the same size"
    
    if np.size(psf1)>np.size(psf2):
        case=True
        big=psf1
        small=psf2
    else:
        big=psf2
        small=psf1
        case=False

    Nb,Mb=big.shape
    Ns,Ms=small.shape
    
    center = Nb/2
    small_side = Ns/2
    
    new_small=np.zeros(big.shape)
    new_small[center-small_side:center+small_side+1,center-small_side:center+small_side+1]=small
    
    if case==True:
        return psf1,new_small
    else:
        return new_small,psf2

def barycenter(img,segmap):
    """ Compute the barycenter of a galaxy from the image and the segemntation map.
    """
    N,M=img.shape
    XX,YY=np.meshgrid(range(M),range(N))
    gal=abs(img*segmap)
    Y = np.average(XX,weights=gal)
    X = np.average(YY,weights=gal)     
    return X,Y


def ferengi_psf_centre(psf,debug=False):
    "Center the psf image using its light barycenter and not a 2D gaussian fit."
    N,M=psf.shape
    
    assert N==M,'PSF image must be square'
    
    if N%2==0:
        center_psf=np.zeros([N+1,M+1])
    else:
        center_psf=np.zeros([N,M])

##    if debug:
##        print np.amax(psf),np.amin(psf)
##        mpl.imshow(psf);mpl.show()
        

    center_psf[0:N,0:M]=psf
    N1,M1=center_psf.shape

    X,Y=barycenter(center_psf,np.ones(center_psf.shape))
    G2D_model = apmodel.models.Gaussian2D(np.amax(psf),X,Y,3,3)

    fit_data = apmodel.fitting.LevMarLSQFitter()
    X,Y=np.meshgrid(np.arange(N1),np.arange(M1))
    with warnings.catch_warnings(record=True) as w:
        pars = fit_data(G2D_model, X, Y, center_psf)

    if len(w)==0:
        cenY = pars.x_mean.value
        cenX = pars.y_mean.value
    else:
        for warn in w:
            print warn
        cenX = center_psf.shape[0]/2+N%2
        cenY = center_psf.shape[1]/2+N%2
    
    dx =(cenX-center_psf.shape[0]/2)
    dy =(cenY-center_psf.shape[1]/2)
    
    center_psf= scndi.shift(center_psf,[-dx,-dy])
    
    
    return center_psf#,pars
    

def ferengi_deconvolve(wide,narrow):#TBD
    "Images should have the same size. PSFs must be centered (odd pixel numbers) and normalized."

    Nn,Mn=narrow.shape #Assumes narrow and wide have the same shape
    
    
    smax = max(Nn,Mn) 
    bigsz=2    
    while bigsz<smax:
        bigsz*=2

    if bigsz>2048:
        print 'Requested PSF array is larger than 2x2k!'
    
    psf_n_2k = np.zeros([bigsz,bigsz],dtype=np.double)
    psf_w_2k = np.zeros([bigsz,bigsz],dtype=np.double)
    
    psf_n_2k[0:Nn,0:Mn]=narrow
    psf_w_2k[0:Nn,0:Mn]=wide
    
#    fig,ax=mpl.subplots(1,2,sharex=True,sharey=True)
#    ax[0].imshow(psf_n_2k)
#    ax[1].imshow(psf_w_2k)
#    mpl.show()

    psf_n_2k=psf_n_2k.astype(np.complex_)
    psf_w_2k=psf_w_2k.astype(np.complex_)
    fft_n = np.fft.fft2(psf_n_2k)
    fft_w = np.fft.fft2(psf_w_2k)
    
    del psf_n_2k,psf_w_2k
    fft_n = np.absolute(fft_n)/(np.absolute(fft_n)+0.000000001)*fft_n
    fft_w = np.absolute(fft_w)/(np.absolute(fft_w)+0.000000001)*fft_w
    
    psf_ratio = fft_w/fft_n

    del fft_n,fft_w
    
#    Create Transformation PSF
    psf_intermed = np.real(np.fft.fft2(psf_ratio))
    psf_corr = np.zeros(narrow.shape,dtype=np.double)
    lo = bigsz-Nn/2
    hi=Nn/2
    psf_corr[0:hi,0:hi]=psf_intermed[lo:bigsz,lo:bigsz]    
    psf_corr[hi:Nn-1,0:hi]=psf_intermed[0:hi,lo:bigsz]    
    psf_corr[hi:Nn-1,hi:Nn-1]=psf_intermed[0:hi,0:hi]    
    psf_corr[0:hi,hi:Nn-1]=psf_intermed[lo:bigsz,0:hi]        
    del psf_intermed
    
    psf_corr = np.rot90(psf_corr,2)
    return psf_corr/np.sum(psf_corr)
    
def ferengi_clip_edge(image,auto_frac=2,clip_also=None,norm=False):#TBD
    N,M=image.shape
    rx = int(N/2/auto_frac)
    ry = int(M/2/auto_frac)
    
    sig=np.array([])
    r=np.array([])
    while True:
        idx = edge_index(image,rx,ry)
        if np.size(idx[0])==0:
            break
        med,sigma,nrej=resistent_mean(image,3)
        sigma*=np.sqrt(np.size(image)-1-nrej)
        sig=np.append(sig,sigma)
        r=np.append(r,rx)
        rx+=1
        ry+=1
    
    new_med,new_sig,new_nrej=resistent_mean(sig,3)
    new_sig*=np.sqrt(np.size(sig)-1-new_nrej)
    
    i=np.where(sig>=new_med*10*new_sig)
    if np.size(i)>0:
        lim = np.min(r[i])
        if np.size(i)>new_nrej*3:
            print 'Large gap?'
        npix = round(N/2.0-lim)
        
        if clip_also is not None:
            clip_also = clip_also[npix:N-1-npix,npix:M-1-npix]
        image=image[npix:N-1-npix,npix:M-1-npix]
    
    if norm==True:
        image/=np.sum(image)
        if clip_also is not None:
            clip_also/=np.sum(clip_also)
    
    if clip_also is not None:
        return npix,image,clip_also
    else:
        return npix,image

def rebin2d(img,Nout,Mout,flux_scale=False):
    """Special case of non-integer magnification for 2D arrays
    from FREBIN of IDL Astrolib.
    """

    N,M = img.shape

    xbox = N/float(Nout)
    ybox = M/float(Mout)

    temp_y = np.zeros([N,Mout])

    for i in range(Mout):
        rstart = i*ybox
        istart = int(rstart)

        rstop = rstart + ybox
        if int(rstop) > M-1:
            istop = M-1
        else:
            istop = int(rstop)

        frac1 = rstart-istart
        frac2 = 1.0 - (rstop-istop)       
        if istart == istop:
            temp_y[:,i] = (1.0-frac1-frac2)*img[:,istart]
        else:
            temp_y[:,i] = np.sum(img[:,istart:istop+1],1) - frac1 * img[:,istart] - frac2 * img[:,istop]

    temp_y = temp_y.transpose()
    img_bin = np.zeros([Mout,Nout])

    for i in range(Nout):
        rstart = i*xbox
        istart = int(rstart)

        rstop = rstart + xbox
        if int(rstop) > N-1:
            istop = N-1
        else:
            istop = int(rstop)

        frac1 = rstart-istart
        frac2 = 1.0 - (rstop-istop)

        if istart == istop:
            img_bin[:,i] = (1.0-frac1-frac2)*temp_y[:,istart]
        else:
            img_bin[:,i] = np.sum(temp_y[:,istart:istop+1],1) - frac1 * temp_y[:,istart]- frac2 * temp_y[:,istop]        

    if flux_scale:
        return img_bin.transpose()
    else:
        return img_bin.transpose()/(xbox*ybox)
        
        
def lum_evolution(zlow,zhigh):
    "Defined Luminosity evolution from L* of Sobral et al. 2013."
    def luminosity(z):
        logL = 0.45*z+41.87
        return 10**(logL)
    return luminosity(zhigh)/luminosity(zlow)

def ferengi_downscale(image_low,z_low,z_high,pix_low,pix_hi,upscale=False,nofluxscale=False,evo=None):

    da_in = cosmos.angular_distance(z_low)
    da_out = cosmos.angular_distance(z_high)
    
    dl_in=da_in*(1+z_low)**2#cosmos.luminosity_distance(z_low)
    dl_out=da_out*(1+z_high)**2#cosmos.luminosity_distance(z_high)
    
    if evo is not None:
##        evo_fact = evo(z_low,z_high)
        evo_fact = evo(0.0,z_high) ### UPDATED TO MATCH FERENGI ALGORITHM
    else:
        evo_fact=1.0
        
    mag_factor = (da_in/da_out)*(pix_low/pix_hi)
    if upscale == True:
        mag_factor=1./mag_factor
    
##    lum_factor = (dl_in/dl_out)**2
    lum_factor = (dl_in/dl_out)**2*(1.+z_high)/(1.+z_low) ### UPDATED TO MATCH FERENGI ALGORITHM
        
    if nofluxscale==True:
        lum_factor=1.0
        
    N,M = image_low.shape

    N_out = int(round(N*mag_factor))
    M_out = int(round(M*mag_factor))

    img_out = rebin2d(image_low,N_out,M_out,flux_scale=True)*lum_factor*evo_fact

    return img_out


def ferengi_odd_n_square():
    #TBD : in principle avoidable if PSF already square image
    #feregi_psf_centre already includes number of odd pixels
    raise NotImplementedError('In principle avoidable if PSF already square image')
    return


def ferengi_transformation_psf(psf_low,psf_high,z_low,z_high,pix_low,pix_high,same_size=None):
    """ Compute the transformation psf. Psf_low and psf_high are the low and high redshift PSFs respectively.
    Also needed as input paramenters the redshifts (low and high) and pixelscales (low and high).
    """    
    
    psf_l = ferengi_psf_centre(psf_low)
    psf_h = ferengi_psf_centre(psf_high)

    da_in = cosmos.angular_distance(z_low)
    da_out = cosmos.angular_distance(z_high)   
    
    N,M=psf_l.shape
    add=0
    
    out_size = round((da_in/da_out)*(pix_low/pix_high)*(N+add))
    
    while out_size%2==0:
        add+=2
        psf_l=np.pad(psf_l,1,mode='constant')
        out_size = round((da_in/da_out)*(pix_low/pix_high)*(N+add))
        if add>N*3:
            return -99
##            raise ValueError('Enlarging PSF failed!')
    
    psf_l = ferengi_downscale(psf_l,z_low,z_high,pix_low,pix_high,nofluxscale=True)
    psf_l = ferengi_psf_centre(psf_l)
    

# Make the psfs the same size (then center)
    psf_l,psf_h=ferengi_make_psf_same(psf_l,psf_h)  


    psf_l = ferengi_psf_centre(psf_l)
    psf_h = ferengi_psf_centre(psf_h)
    

# NORMALIZATION    
    psf_l/=np.sum(psf_l)
    psf_h/=np.sum(psf_h)

    return psf_l,psf_h,ferengi_psf_centre(ferengi_deconvolve(psf_h,psf_l))

def ferengi_convolve_plus_noise(image,psf,sky,exptime,nonoise=False,border_clip=None,extend=False):
    
    if border_clip is not None: ## CLIP THE PSF BORDERS (CHANGE IN INPUT PSF)
        Npsf,Mpsf=psf.shape
        psf = psf[border_clip:Npsf-border_clip,border_clip:Mpsf-border_clip]
    
    Npsf,Mpsf=psf.shape
    Nimg,Mimg=image.shape

    out = np.pad(image,Npsf,mode='constant') #enlarging the image for convolution by zero-padding the psf image size
    
    out = apcon.convolve_fft(out,psf/np.sum(psf))
    
    Nout,Mout=out.shape
    if extend==False:
        out=out[Npsf:Nout-Npsf,Mpsf:Mout-Mpsf]
        
    Nout,Mout=out.shape # grab new dimensions, if escaped
    
    Nsky,Msky=sky.shape

    if nonoise==False:
        ef= Nout%2
        try:
            out+=sky[Nsky/2-Nout/2:Nsky/2+Nout/2+ef,Msky/2-Mout/2:Msky/2+Mout/2+ef]+\
                 np.sqrt(np.abs(out*exptime))*npr.normal(size=out.shape)/exptime
        except ValueError:
##            raise ValueError('Sky Image not big enough!')
            return -99*np.ones(out.shape)
        
    return out

def dump_results(image,psf,imgname_in,bgimage_in,names_out,lowz_info,highz_info):
    name_imout,name_psfout = names_out

    Hprim = pyfits.PrimaryHDU(data=image)
    hdu = pyfits.HDUList([Hprim])
    hdr_img = hdu[0].header
    hdr_img['INPUT']=imgname_in
    hdr_img['SKY_IMG']=bgimage_in
    for key in lowz_info.keys():
        hdr_img['%s_i'%key[:4]]=(lowz_info[key],'%s value for input lowz object'%(key))
    hdr_img['comment']='Using ferengi.py version %s'%version
    for key in highz_info.keys():
        hdr_img['%s_o'%key[:4]]=(highz_info[key],'%s value for input highz object'%(key))
    hdu.writeto(name_imout,clobber=True)
    pyfits.writeto(name_psfout,psf,clobber=True)
    return
    
def ferengi(imgname,background,lowz_info,highz_info,namesout,imerr=None,noflux=False,evo=None,noconv=False,kcorrect=False,extend=False,nonoise=False,border_clip=3):

    Pl=pyfits.getdata(lowz_info['psf'])
    Ph=pyfits.getdata(highz_info['psf'])
    sky=pyfits.getdata(background)
    image=pyfits.getdata(imgname)
    if imerr is None:
        imerr=1/np.sqrt(np.abs(image))
    else:
        imerr=pyfits.getdata(imerr)
    
    if kcorrect:
        raise NotImplementedError('K-corrections are not implemented yet')
    else:
        img_nok = maggies2cts(cts2maggies(image,lowz_info['exptime'],lowz_info['zp']),highz_info['exptime'],highz_info['zp'])#*1000.
        img_downscale = ferengi_downscale(img_nok,lowz_info['redshift'],highz_info['redshift'],lowz_info['pixscale'],highz_info['pixscale'],evo=evo,nofluxscale=noflux)

        psf_low = Pl
        psf_hi = Ph

    
    median = scndi.median_filter(img_downscale,3)
    idx = np.where(np.isfinite(img_downscale)==False)
    img_downscale[idx]=median[idx]
    
    idx = np.where(img_downscale==0.0)
    img_downscale[idx]=median[idx]
    
    X=img_downscale.shape[0]*0.5
    Y=img_downscale.shape[1]*0.5 ## To be improved
    
    img_downscale-=ring_sky(img_downscale, 50,15,x=X,y=Y,nw=True)
    
    if noconv==True:
        dump_results(img_downscale/highz_info['exptime'],psf_low/np.sum(psf_low),imgname,background,namesout,lowz_info,highz_info)
        return img_downscale/highz_info['exptime'],psf_low/np.sum(psf_low)

    try:
        psf_low,psf_high,psf_t = ferengi_transformation_psf(psf_low,psf_hi,lowz_info['redshift'],highz_info['redshift'],lowz_info['pixscale'],highz_info['pixscale'])
    except TypeError as err:
        print('Enlarging PSF failed! Skipping Galaxy.')
        return -99,-99

    try:
        recon_psf = ferengi_psf_centre(apcon.convolve_fft(psf_low,psf_t))
    except ZeroDivisionError as err:
        print('Reconstrution PSF failed!')
        return -99,-99
##    pyfits.writeto('transform_psf_dopterian.fits',psf_t,clobber=True)
    
    recon_psf/=np.sum(recon_psf)
    
    img_downscale = ferengi_convolve_plus_noise(img_downscale/highz_info['exptime'],psf_t,sky,highz_info['exptime'],nonoise=nonoise,border_clip=border_clip,extend=extend)
    if np.amax(img_downscale) == -99:
        print 'Sky Image not big enough!'
        return -99,-99

##    import matplotlib.pyplot as mpl
##    mpl.imshow(img_downscale);mpl.show()

    dump_results(img_downscale,recon_psf,imgname,background,namesout,lowz_info,highz_info)
    return img_downscale,recon_psf



if __name__=='__main__':
    PlowName='psf_sdss.fits'
    PhighName='psf_acs.fits'
    BgName='sky_ACSTILE_40x40.fits'
    InputImName='galaxy.fits'
    
    lowz_info = {'redshift':0.017,'psf':PlowName,'zp':28.235952,'exptime':53.907456,'filter':'r','lam_eff':6185.0,'pixscale':0.396}
    highz_info = {'redshift':0.06,'psf':PhighName,'zp':25.947,'exptime':6900.,'filter':'f814w','lam_eff':8140.0,'pixscale':0.03}
    
    import time as t
    t0=t.time()
#    imOUT,psfOUT = ferengi(InputImName,BgName,lowz_info,highz_info,['smooth_galpy_evo.fits','smooth_psfpy_evo.fits'],noconv=False,evo=lum_evolution)
    imOUT,psfOUT = ferengi(InputImName,BgName,lowz_info,highz_info,['smooth_galpy.fits','smooth_psfpy.fits'],noconv=False,evo=None)
    print 'elapsed %.6f secs'%(t.time()-t0)

#    fig,ax=mpl.subplots(1,2)
#    ax[0].imshow(imOUT)
#    ax[1].imshow(psfOUT)
#    mpl.show()



from scipy.integrate import simps,quad
import numpy as np

""" COSMOLOGY MODULE
27/08/2013  - Define Angular Diameter Distance and Cosmological parameters
            - Compare function against http://en.wikipedia.org/wiki/File:Angular_diameter_distance.gif
"""
#### PLACNK COSMOLOGY
Obar=0.049                 #Baryon Density Parameter
Omat=0.3#0.3175                    #Matter Density Parameter
Ok=0.0                      #Curvature Density Parameter
Orad=0.0                    #Radiation Density Parameter
Ow=0.7#0.6825                      #Dark Energy Density Parameter
w=-1.0                      #DE equation of state parameter p/rho=w
H0=70.0#67.11                      #km/s/Mpc
Msun=1.989e30               #kg
Mpc=3.0857e19               #km
c=2.9979e5                  #km/s

def Hubble(z,pars=None):
    "Returns the value for the standard Hubble parameter at a redshift z"
    P={'h':H0/100.,'r':Orad,'m':Omat,'k':Ok,'l':Ow,'w':w}
    if not (pars==None):
        for p in pars:
            P[p] = pars[p]
    return 100*P['h']*np.sqrt(P['r']*(1+z)**4.+P['m']*(1+z)**3.+P['k']*(1+z)**2.+P['l']*(1+z)**(3*(1.+P['w'])))


def invHubble(z,pars=None):
    return 1.0/Hubble(z,pars=pars)

def comov_rad(z,pars=None,npoints=10000):
    """Returns the comoving radial distance corresponding to the redshift z in Mpc
    If nedeed, multiply by h to get result in units of (h**-1 Mpc)"""
    radius,err=quad(invHubble,0,z)
    return c*radius

def lookback_time(z,pars=None,npoints=10000):
    "Computes the lookback time (in Gyr) at a redshift z"
    if z==0:
        z=1e-4 
    z_points=np.linspace(1e-5,z,npoints)
    I = 1./((1+z_points)*Hubble(z_points,pars=pars))
    tl = simps(I,z_points)*Mpc/(365.25*24*60*60*1e9)
    return tl

def find_z_tL(time,pars=None):
    from scipy.optimize import newton
    def minim(z):
        return lookback_time(z,pars=None)-time
    res=newton(minim,0.5)
    return res
    
def angular_distance(z,pars=None):
    "Computes the angular diameter distance (Mpc) in a standard LCDM cosmology"
    return comov_rad(z,pars=pars)/(1+z)


def luminosity_distance(z,pars=None):
    "Computes the luminosity distance (Mpc) in a standard LCDM cosmology"
    return comov_rad(z,pars=pars)*(1+z)

def plot_lookback():
    import matplotlib.pyplot as pl
    Zs = np.linspace(1e-5,10,1000)   
    times=np.zeros(len(Zs))
    for i in range(len(Zs)):
        times[i]=lookback_time(Zs[i])
    fig,ax=pl.subplots()
    ax.plot(Zs,times,'k',lw=3)
    ax.hlines(13.8,0,10,'k',':')
    ax.set_xlabel(r'$z$')
    ax.set_ylabel(r'$t_L\ [\mathrm{Gyr}]$')
    aux = fig.add_axes([0.5,0.2,0.3,0.3])
    aux.plot(Zs,times,'k',lw=3)
    aux.set_xlim(2,6)
    aux.set_ylim(10,13)
    aux.grid(True)
    pl.show()
     
def test_distances():
    import matplotlib.pyplot as pl
    Zs = np.linspace(1e-5,10,1000)
    distsA=np.zeros(len(Zs))
    distsA2=np.zeros(len(Zs))
    distsL=np.zeros(len(Zs))
    distsL2=np.zeros(len(Zs))
    for i in range(len(Zs)):
        distsA[i]=angular_distance(Zs[i])
        distsA2[i]=angular_distance(Zs[i],pars={'l':0.0,'m':1.0})
        distsL[i]=luminosity_distance(Zs[i])
        distsL2[i]=luminosity_distance(Zs[i],pars={'l':0.0,'m':1.0})
    
    fig,(ax1,ax2)=pl.subplots(1,2,figsize=(15,8))
    ax1.plot(Zs,distsA,'b-',Zs,distsA2,'r-')
#    ax1.set_xlim(2,5)
    ax1.set_xlabel(r'$z$')
    ax1.set_ylabel(r'$d_A$')
    ax2.plot(Zs,distsL,'b-',Zs,distsL2,'r-')
#    ax2.set_xlim(2,5)
    ax2.set_xlabel(r'$z$')
    ax2.set_ylabel(r'$d_L$')    
    pl.show()
    return

def plot_conversion_arcsec_to_kpc():
    import matplotlib.pyplot as mpl
    Zs = np.linspace(0.1,6.9,1000)

    dists_Planck=np.array([angular_distance(z) for z in Zs])

    kpc_per_arcsec = 1*dists_Planck/(180/np.pi*3600)*1000

    fig,ax=mpl.subplots(figsize=(15,8))

    ax.plot(Zs,1.0/kpc_per_arcsec,'-',color='RoyalBlue',lw=3)
    ax.set_xlabel(r'$z$')
    ax.set_ylabel(r'$S\ [\mathrm{arcsec/kpc}]$')
    ax.minorticks_on()
    ax.set_xlim(0.1,6.5)
    fig.savefig('Scale_arcsec_per_kpc.png')
    mpl.show()

def compare_WMAP_Planck_cosmologies():
    import matplotlib.pyplot as pl
    Zs = np.linspace(0.1,6.9,1000)

    dists_Planck=np.array([angular_distance(z) for z in Zs])
    dists_WMAP=np.array([angular_distance(z,pars={'h':0.7,'m':0.3,'l':0.7}) for z in Zs])


    diff = dists_Planck/dists_WMAP

    fig,ax=pl.subplots(2,1,sharex=True,figsize=(15,13))
    pl.subplots_adjust(hspace=0.0)
    ax[1].plot(Zs,diff,'k-',lw=2)
    ax[1].set_xlabel(r'$z$')
    ax[1].set_ylabel(r'$D_A^\mathrm{Planck}/D_A^\mathrm{STD}$')
    ax[1].hlines(1,0,7,'r','--',alpha=0.5,lw=2)
    ax[1].set_xlim(0.1,6.9)
    ax[1].set_ylim(0.985,1.045)

    ax[0].plot(Zs,dists_WMAP,'r-',lw=2,label=r'STD: $H_0=70\mathrm{km\ s^{-1}Mpc^{-1}}; \Omega_m=0.3; \Omega_\Lambda=0.7$')
    ax[0].plot(Zs,dists_Planck,'g-',lw=2,label='Planck')
    ax[0].set_ylabel(r'$D_A\ [\mathrm{Mpc}]$')
    ax[0].legend(loc='lower right')
    ax[0].set_ylim(250,1900)
    fig.savefig('Impact_Cosmology_Choice.png')
    pl.show()

    
if __name__=='__main__':
#    test_distances()
#    plot_lookback()
#    compare_WMAP_Planck_cosmologies()
    plot_conversion_arcsec_to_kpc()
